/*
  Charger for Medical batteries
  Board: Arduino Nano

  V4 - Modified pin assignments to suit "final" layout
  V5 - Changed I2C code to use Wire library

*/

#include <Wire.h>
#include <LiquidCrystal.h>

#define ADDRESS 0x0b


// Some Pin Allocations

#define CHARGE    11
#define DISCHARGE 12
#define OFF       0
#define ON        1

// LCD:
const int  rs = 4, en = 5, d4 = 6, d5 = 7, d6 = 8, d7 = 9;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);

char fp_thing[21];

unsigned long ticker, next_activity, last_activity; // used for 10 second i2c & activity timer

unsigned int fn, old_fn;  // Currently selected charger function

int Amps;
float Amps2;
float Amps3;
float volts;
unsigned int soc;
float temp;
int bv;
float amp_hours = 0;

// Global battery info

uint16_t mode = 0;
uint16_t stat = 0;

// Global State Machine data
// Used in capacity test function

unsigned int G_state = 0;
unsigned long q1 = 0, q0 = 0;   // Charge counter - extra long
#define FLAT_BATTERY 705   //A2D count equiv to 10.5 volts

// ---------------------------------------------------------------------

int read_word(int reg) {

  int b1, b2;

  Wire.beginTransmission(ADDRESS);
  Wire.write(byte(reg));
  Wire.endTransmission();

  Wire.requestFrom(ADDRESS, 2);

  if (Wire.available() == 2 ) { // only if two bytes were received

    b1 = Wire.read();
    b2 = Wire.read();

  }

  Wire.endTransmission();

  return ((b2 << 8) | b1);
}


//---------------------------------------------------------------------------------------------

void  report_full_charge_capacity(void) {

  float Capacity = (float)read_word(0x18);

  Serial.print("Full Charge Capacity = ");

  if ( mode & 0x8000 ) {
    //    Serial.println("  Capacity in 10mWh");
    Capacity = Capacity / 100; // convert to Wh
    Serial.print(  Capacity, 3);
    Serial.println("Wh");
  }
  else
  {
    //    Serial.println("  Capacity in mAh");
    Capacity = Capacity / 1000; // convert to Ah
    Serial.print(  Capacity, 3);
    Serial.println("Ah");
  }

}

//---------------------------------------------------------------------------------------------

void  report_design_capacity(void) {

  float Capacity = (float)read_word(0x18);

  Serial.print("Design Capacity = ");

  if ( mode & 0x8000 ) {
    //    Serial.println("  Capacity in 10mWh");
    Capacity = Capacity / 100; // convert to Wh
    Serial.print(  Capacity, 3);
    Serial.println("Wh");
  }
  else
  {
    //    Serial.println("  Capacity in mAh");
    Capacity = Capacity / 1000; // convert to Ah
    Serial.print(  Capacity, 3);
    Serial.println("Ah");
  }

}

//---------------------------------------------------------------------------------------------

float  NEW_request_volts(void) {

  int count = 5;
  float volts;

  while ( count >= 0 ) {
    volts = (float)read_word(0x15) / 1000;
    if ( (volts <= 13) && (volts >= 10) ) return (volts);
    count--;
    Serial.print("x"); //Debugging
  }

  return (666);

}

//---------------------------------------------------------------------------------------------

float  NEW_request_amps(void) {

  int count = 5;
  float amps;

  while ( count >= 0 ) {
    amps = (float)read_word(0x14) / 1000; //P41
    if ( (amps <= 10) && (amps >= 0) ) return (amps);
    count--;
    Serial.print("x"); //Debugging
  }

  return (666);

}

//---------------------------------------------------------------------------------------------

void  report_amps(void) {

  int Amps = (int)read_word(0x0a); // NB signed integer mA
  float Amps2 = (float)Amps / 1000;
  Serial.print("Current = ");
  Serial.print( Amps2, 3);
  Serial.println("A");
}

//---------------------------------------------------------------------------------------------

float  NEW_report_temp(void) {

  int count = 5;
  float temp;

  while ( count >= 0 ) {
    temp = (float)read_word(0x08) / 10.0 - 273.0;
    if ( (temp <= 100) && (temp > -20) ) return (temp);
    count--;
    Serial.print("x"); //Debugging
  }
  return (666);
}

//--------------------------------------------------------------------------------------------

unsigned int  NEW_report_rel_soc(void) {

  int count = 5;
  unsigned int soc;

  while ( count >= 0 ) {
    soc = (unsigned int)read_word(0x0d);
    if ( (soc <= 100) ) return (soc);
    count--;
    Serial.print("x"); //Debugging
  }
  return (666);

}

//---------------------------------------------------------------------------------------------

void  report_cycle_count(void) {

  unsigned int count = (unsigned int)read_word(0x17);
  Serial.print("Cycle count = ");
  Serial.println( count);

}

//---------------------------------------------------------------------------------------------

void  report_resistance(void) {
  float V1, V2, Amps2, R;
  int Amps;

  V1 = (float)read_word(0x09) / 1000;
  digitalWrite(4, HIGH);
  delay(1000);
  V2 = (float)read_word(0x09) / 1000;
  delay(1000);
  Amps = (int)read_word(0x0a); // NB signed integer mA
  digitalWrite(4, LOW);

  Amps2 = (float)Amps / 1000;

  Serial.print("V1 = ");
  Serial.print( V1 , 3);
  Serial.println(" V");

  Serial.print("V2 = ");
  Serial.print( V2 , 3);
  Serial.println(" V");

  Serial.print("A = ");
  Serial.print( Amps2 , 3);
  Serial.println(" A");

  R = (V2 - V1) / Amps2;

  Serial.print("Internal Resistance = ");
  Serial.print( R , 3);
  Serial.println(" Ohm");

}



//---------------------------------------------------------------------------------------------

void  report_battery_mode(void) {

  // mode is a global variable

  mode = (uint16_t)read_word(0x03);

  //  Serial.println("WARNING. Following data is probably wrong. Debugging is incomplete.");
  Serial.print("Mode   word = ");
  printword(mode);
  //  Serial.println( mode, BIN);

  if ( mode & 0x8000 ) Serial.println("  Capacity in 10mW or 10mWh"); else Serial.println("  Capacity in mA or mAh");
  if ( mode & 0x4000 ) Serial.println("  Disable V&I broadcasts");    else Serial.println("  Enable V&I broadcasts");
  if ( mode & 0x2000 ) Serial.println("  Disable Alarm broadcasts");  else Serial.println("  Enable Alarm broadcasts");

  if ( mode & 0x0200 ) Serial.println("  Primary battery");           else Serial.println("  Secondary battery");
  if ( mode & 0x0100 ) Serial.println("  Internal CC enabled");       else Serial.println("  Internal CC disabled");

  if ( mode & 0x0080 ) Serial.println("  Conditioning cycle required");     else Serial.println("  Condition is OK");

  if ( mode & 0x0020 ) Serial.println("  P/S supported");             else Serial.println("  P/S not supported");
  if ( mode & 0x0001 ) Serial.println("  Has internal CC");           else Serial.println("  Has no internal CC");

  Serial.println();
}

//---------------------------------------------------------------------------------------------

void  report_battery_stat(void) {

  // stat is a global variable

  stat = (uint16_t)read_word(0x16);

  //  Serial.println("WARNING. Following data is probably wrong. Debugging is incomplete.");
  Serial.print("Status word = ");
  printword(stat);
  //  Serial.println( stat, BIN);

  if ( stat & 0x8000 ) Serial.println("  Over Charged ALARM");
  if ( stat & 0x4000 ) Serial.println("  Terminate Charge ALARM");
  if ( stat & 0x1000 ) Serial.println("  Over Temperature ALARM");

  if ( stat & 0x0800 ) Serial.println("  Terminate discharge ALARM");
  if ( stat & 0x0200 ) Serial.println("  Reamaining Capacity ALARM");
  if ( stat & 0x0100 ) Serial.println("  Remaining time ALARM");

  if ( stat & 0x0080 ) Serial.println("  Newly initialised");
  if ( stat & 0x0040 ) Serial.println("  Discharging");
  if ( stat & 0x0020 ) Serial.println("  Fully charged");
  if ( stat & 0x0010 ) Serial.println("  fully discharged");

  Serial.print("  ERROR Code = "); Serial.println( stat & 0x0f , HEX);

  Serial.println();
}

//---------------------------------------------------------------------------------------------

void printword(uint16_t x) {

  for ( int i = 1; i <= 16; i++ ) {

    if ( x & 0x8000 ) Serial.print("1"); else Serial.print("0");
    if ( !(i % 4) ) Serial.print(" ");
    x = x << 1;

  }
  Serial.println();
}

//---------------------------------------------------------------------------------------------

int function( void ) {

  int f;

  f = analogRead(A1);

  if ( f < 163 ) return (3);

  if ( f < 501 ) return (2);

  if ( f < 860 ) return (1);

  return (0);

}
//---------------------------------------------------------------------------------------------

void setup() {

  Wire.begin();                   // join i2c bus
  Serial.begin(9600);
  Serial.println("Set Up");

  lcd.begin(24, 2);               // set up the LCD's number of columns and rows:

  pinMode(CHARGE, OUTPUT);        // Charging control
  digitalWrite(CHARGE, OFF);      // Off

  pinMode(DISCHARGE, OUTPUT);     // Discharging control
  digitalWrite(DISCHARGE, OFF);   // Off

  fn = function();                // Switch function
  old_fn = fn;

  delay(1000);

  ticker = millis();              // i2c activity timer
  next_activity = ticker + 100;   // schedule earliest time of next i2c action
  last_activity = ticker;

  set_charging_voltage();
  Serial.println("END of Set Up");

}

//---------------------------------------------------------------------------------------------

void loop() {

  fn = function();

  if ( fn != old_fn ) { // we have detected switch fiddling

    old_fn = fn;

    ticker = millis();              // i2c activity timer
    next_activity = ticker + 500;   // schedule earliest time of next i2c action
    last_activity = ticker;

  }

  ticker = millis();              // i2c activity timer

  if ( ticker >= next_activity ) { // time for some i2c (& other) action
    next_activity = ticker + 10000;   // 10 seconds to next. Begin current session now.

    Serial.println("*");

    Wire.beginTransmission(ADDRESS);
    if ( Wire.endTransmission() == 0 ) {

      volts = NEW_request_volts();
      set_charging_voltage(); // Set Charging Voltage

      // Some serial debugging stuff
      // Requested Charge voltage
      Serial.print("Battery found. Requested Charge Voltage = ");
      Serial.print( volts , 3);
      Serial.println("V");
      Serial.println();

      soc = NEW_report_rel_soc();
      temp = NEW_report_temp();
      Amps = (int)read_word(0x0a); // NB signed integer mA
      Amps2 = (float)Amps / 1000;
      Amps3 = NEW_request_amps();


      switch ( fn ) {

        case 0:   // Just report battery readings

          Serial.println("case 0. Display parameters.");

          G_state = 0;
          digitalWrite(CHARGE, OFF);
          digitalWrite(DISCHARGE, OFF);

          delay(500);
          LCD_top_line();
          LCD_bottom_line_1();

          break;

        case 1:   // Charge to 100%

          Serial.println("case 1. Charge to 100%");

          G_state = 0;
          digitalWrite(DISCHARGE, OFF);

          if ( ( Amps3 > 0.010 ) && ( soc < 100 ) ) {
            digitalWrite(CHARGE, ON);
            Serial.println("Charging");
          } else {
            digitalWrite(CHARGE, OFF);
            Serial.println("Charge terminated");
          }

          delay(500);
          LCD_top_line();
          LCD_bottom_line_1();

          break;

        case 2:   // Charge or Discharge to 50...52%

          Serial.println("case 2. Charge or Discharge to ~50%");

          G_state = 0;

          if ( soc < 50 ) {
            digitalWrite(DISCHARGE, OFF);
            digitalWrite(CHARGE, ON);
            Serial.println("Charging to >= 50%");

          } else {

            digitalWrite(CHARGE, OFF);

            if ( soc > 52 ) {
              digitalWrite(DISCHARGE, ON);
              Serial.println("Discharging to <= 52%");
            } else {
              digitalWrite(DISCHARGE, OFF);
              Serial.println("Charge/Discharge completed");
            }

          }

          delay(500);
          LCD_top_line();
          LCD_bottom_line_1();

          break;

        case 3:   // Measure Battery capacity

          Serial.println("case 3. Measure Battery capacity");

          switch ( G_state ) {

            case 0:   // Not yet started

              digitalWrite(DISCHARGE, OFF);
              digitalWrite(CHARGE, OFF);

              Serial.println("Ah test, state=0");
              q1 = 0, q0 = 0;  // Initialise Ah counters
              G_state = 1;    // Start

              Serial.print("G_state =");
              Serial.println(G_state);

              //              LCD_top_line();
              //              LCD_bottom_line_2();

              break;

            case 1:   // Charge to 100%

              Serial.println("Ah test, state=1. Charge to 100%");

              //             digitalWrite(DISCHARGE, OFF);

              if ( ( Amps3 > 0.010 ) && ( soc < 100 ) ) {

                digitalWrite(CHARGE, ON);
                Serial.println("Charging to 100%");

                G_state = 1;

              } else {

                digitalWrite(CHARGE, OFF);    // Prepare for accumulation
                digitalWrite(DISCHARGE, ON);
                Serial.println("100% Charge reached");

                ticker = millis();              // i2c activity timer

                last_activity = ticker;

                G_state = 2;
              }

              Serial.print("G_state =");
              Serial.println(G_state);

              break;

            case 2:   // Discharge to "flat", whatever that has to be defined as.

              Serial.println("Ah test, state=2. Discharge to 0%");

              bv = analogRead(A2);

              ticker = millis();              // i2c activity timer

              q0 += bv * (ticker - last_activity);

              while ( q0 >= 1e6 ) {
                q0 -= 1e6;
                q1++;
                Serial.print(" q1 = ");
                Serial.println( q1 );
                Serial.print(" q0 = ");
                Serial.println( q0 );

              }

              amp_hours = (q1 * 1e6 + q0) / 3.632e9 ;

              Serial.print(" Amp_hours = ");
              Serial.println( amp_hours, 3 );
              //              LCD_top_line();
              //              LCD_bottom_line_2();

              last_activity = ticker; // Ready for next interval

              if ( ( bv <= FLAT_BATTERY ) || ( soc < 2 ) ) {  // Are we flat yet? Check this!

                digitalWrite(CHARGE, OFF);    // Just to be safe
                digitalWrite(DISCHARGE, OFF);    // End accumulation

                G_state = 3;

              }

              Serial.print("G_state =");
              Serial.println(G_state);
              break;

            case 3:   // Charge to ~50"

              Serial.println("Ah test, state=3. Charge to 50%");
              if ( soc < 50 ) {
                digitalWrite(DISCHARGE, OFF);
                digitalWrite(CHARGE, ON);

              } else {

                digitalWrite(CHARGE, OFF);

                if ( soc > 52 )
                  digitalWrite(DISCHARGE, ON);
                else {
                  digitalWrite(DISCHARGE, OFF);
                  G_state = 4;
                }
              }

              Serial.print("G_state =");
              Serial.println(G_state);
              break;

            case 4:
            default:

              digitalWrite(DISCHARGE, OFF);   // To be sure, to be sure
              digitalWrite(CHARGE, OFF);

              Serial.print("G_state =");
              Serial.println(G_state);
              break;

          }   // End of switch - capacity measurement state machine

          delay(500);
          LCD_top_line();
          LCD_bottom_line_2();

          break;

        default:    // Should never happen!
          Serial.println("Oops");
          break;

      }   // End of Switch - function selection.

    } else {

      // No battery connected

      digitalWrite(CHARGE, OFF);
      digitalWrite(DISCHARGE, OFF);
      G_state = 0;

      analogWrite(10, 128); // Test value
      Serial.println("No battery found");
      lcd.setCursor(0, 0);
      lcd.print("  Battery not Detected  ");
      lcd.setCursor(0, 1);
      lcd.print("                        ");

    }

    Serial.println();

    int vd;
    float va, r;

    vd = analogRead(A0);

    //    if ( vd < 200 ) vd += 2;  // fiddle factor

    va = (float)vd;
    r = (4700 * va) / (1023.0 - va);

    Serial.print(" Analog read A0 = ");
    Serial.print( vd );
    Serial.print(" Rt = ");
    Serial.print(  r, 0 );
    Serial.println(" Ohm" );


    Serial.print(" Analog read A1 = ");
    Serial.print( analogRead(A1) );
    Serial.print(" Function = ");
    Serial.println(  function() );

    Serial.print(" Analog read A2 = ");
    Serial.println( analogRead(A2) );

    Serial.print(" Analog read A3 = ");
    Serial.println( analogRead(A3) );

    //  A4 NOW USED BY I2C
    
    //Serial.print(" Analog read A4 = ");
    //Serial.println( analogRead(A4) );

    //  A5 now used by I2C

    //  Serial.print(" Analog read A5 = ");
    //  Serial.println( analogRead(A5) );

    Serial.print(" Analog read A6 = ");
    Serial.println( analogRead(A6) );

    Serial.print(" Analog read A7 = ");
    Serial.println( analogRead(A7) );

    Serial.println();

  }
}

void set_charging_voltage( void ) {
  int v_trim;

  // Set Charging Voltage
  // Measured sensitivity 133 counts == 1.0 volt change
  // Regulator "centre" voltage = 12.45, set with no battery

  v_trim = (int)( ( NEW_request_volts() - 11.49) * 133 );
  if ( v_trim < 0 ) v_trim = 0;
  if ( v_trim > 255 ) v_trim = 255;
  analogWrite(10, v_trim); // Adjusts main regulator


}

void LCD_top_line( void ) {  // Do Top line of LCD

  // Temperature
  Serial.print(" Temperature = ");
  Serial.print( temp , 1);
  Serial.println("C");

  lcd.setCursor(0, 0);
  lcd.print("T = ");
  dtostrf(temp, 5, 1, fp_thing);
  lcd.print(fp_thing);
  lcd.print("C  ");

  // State of Charge
  Serial.print(" State of charge = ");
  Serial.print( soc);
  Serial.println("%");

  lcd.setCursor(12, 0);
  lcd.print("SOC = ");
  lcd.print(soc);
  lcd.print("%  ");
  // End of Top Line
}

void LCD_bottom_line_1( void ) {

  // Some serial debugging stuff
  // Current
  Serial.print(" Current = ");
  Serial.print( Amps2, 3);
  Serial.println("A");

  // Requested Charge Current
  Serial.print(" Requested Charge Current = ");
  Serial.print( Amps3, 3);
  Serial.println("A");

  // Display inst Amps & Requested charging Amps.
  lcd.setCursor(0, 1);
  lcd.print("A = ");
  dtostrf(Amps2, 6, 3, fp_thing);
  lcd.print(fp_thing);

  lcd.setCursor(12, 1);
  lcd.print("Arq = ");
  dtostrf(Amps3, 6, 3, fp_thing);
  lcd.print(fp_thing);

}

void LCD_bottom_line_2( void ) {

  // Capacity test. Display Ah
  lcd.setCursor(0, 1);
  lcd.print("Capacity = ");
  dtostrf(amp_hours, 6, 3, fp_thing);
  lcd.print(fp_thing);
  lcd.print("Ah        ");

}
